use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007000
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib  ) );
__DATA__
ABCDEFGHIJKLMNOPQRSTUVWXYZ
AWS
AnnoCPAN
CLI
CPAN
Cas
CharacterGroupNotDefinedIronPolicyException
Github
HTTPS
IronCaches
IronHTTPCallException
IronIO
IronPolicyException
Ironcache
JSON
MERCHANTABILITY
NoIronPolicyException
OAuth
POSIX
Params
RESTful
Rackspace
SaaS
Storable
Subdirectory
TODO
Timestamp
YAML
abcdefghijklmnopqrstuvwxyz
alnum
ascii
aws
cas
dir
filename
http
https
io
ironcache
ironmq
ironworker
json
licensable
lv
msg
multi
params
runtime
scalability
semafores
startup
subitem
subparam
succcessful
successfull
timestamp
tradename
unreserves
url
webhooks
webservices
wildcard
wildcards
Mikko
Koivunalho
mikko
lib
IO
Iron
ClientBase
Common
Connection
Connector
ConnectorBase
IronCache
Api
Cache
Client
Item
Policy
IronMQ
Message
Queue
IronWorker
Task
PolicyBase
CharacterGroup
